--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   LPSolve.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description: 
--   The lp_solve front end.
--   See http://lpsolve.sourceforge.net/5.5/.
--
--   RAML expects the binary lp_solve to be in the path.
--   You can give the path to the binary by changing the variable lpSolveBin.
--
--   The module provides two functions.
--
--   The function solve takes a list of constraints and a objective function
--   (and a path for the temp files), writes the constraints and the objective
--   function to a file, runs lp_solve on the file, and parses the solution
--   into a variable mapping (or throws an error if the constraint set could
--   not be solved).
--
--   The function lpSolveResult is a parser for the lp_solve output. 


module LPSolve(solve) where

import qualified System.Cmd as Cmd     -- to execute shell commands
import qualified Data.Map as M
import Text.ParserCombinators.Parsec   -- to parse the output
import System.IO                       -- to work with files
import qualified GHC.IO.Exception as Exception

import RAMLTypes
import Util


-- The path to the lp_solve binary.
lpSolveBin = "lp_solve"


solve :: (Monad err) => TempFilesFlag -> Constraints -> [Objective] -> FilePath -> IO (err (M.Map CVarId Double))
-- The function solve takes a list of constraints and a objective function
-- (and a path for the temp files), writes the constraints and the objective
-- function to a file, runs lp_solve on the file, and parses the solution
-- into a variable mapping (or throws an error if the constraint set could
-- not be solved).

-- Haskell provides a nice way to create temp files.  We create
-- two in the temp dir: constrXXX.lp and solutionXXX.lp. Where XXX is a
-- random number.  We only need the file name and close the handles right away.

-- Then store the linar program, run lp_solve and store the output, 
-- read the solution from the output file, and parse the solution.
solve tempFilesFlag constraints objectives tempDir =
    do { (constrFileName, constrFile) <- openTempFile tempDir "constr.lp"        
       ; (solutionFileName, solutionFile) <- openTempFile tempDir "solution.lp"  
       ; hClose constrFile                                                       
       ; hClose solutionFile                                                     
       ; writeFile constrFileName lpStr                                                     
       ; Cmd.system $ lpSolveBin ++ " " ++ constrFileName ++ " > " ++ solutionFileName       
       ; solutionString <- readFile solutionFileName                                         
       ; let solution = case parse lpSolveResult solutionFileName solutionString of
                          Left err -> fail ("lp_solve finished unsuccessful. " ++ (show err))
                          Right x -> return x
       ; case tempFilesFlag of 
            DeleteFiles -> Cmd.system $ "rm " ++ constrFileName ++ " " ++ solutionFileName
            KeepFiles -> return Exception.ExitSuccess
       ; return solution
       }
   where                                   -- The lp_solve input:
      lpStr = "min: "                      -- the objective function should be minimized & starts hence with 'min:'
         ++ (showWithSep " + " objectives) -- followed by the objectives of the form q1 x1  + .... + q2 x2
         ++ ";\n"                          -- ended with ';'
         ++ "\n"                           -- The rest of the input consists of
         ++ (showWithSep ";\n" constraints)-- the constraints (e.g. "0.5 x < 1.2 y + z + 3.0") separated with ';'
         ++ ";"                            -- ended with ';'




-- The parser for the lp_solve output.
-- The output can start with some empty lines.
-- Then there is ether the line "This problem is infeasible" (unsolvable)
-- or the line "Actual values of the variables:" (variableValues).
-- In the first we raise an error.
-- In the second case we collect the values of the variables.
lpSolveResult = unsolvable
                <|> variableValues
                <|> do { skipLine
                       ; lpSolveResult
                       }
   where
      unsolvable = do { string "This problem is infeasible"
                      ; fail $ "The linear program is infeasible."
                      }
      
      variableValues = do { string "Actual values of the variables:"
                          ; spaces
                          ; map <- many varValue
                          ; return $ M.fromList map
                          }
      
      varValue  = do { varId <- varName
                     ; spaces
                     ; value <- double
                     ; spaces
                     ; return (CVarId $ cVarFromString varId,value)
                     }
      
      
      varName = do { l  <- letter -- Variable names as defined by the lp_solve manual.
                   ; ls <- many ( alphaNum <|> oneOf "_[]{}/.&#$%~'@^" )
                   ; return $ l:ls
                   }


      double = do { ds <- many digit
                  ; do { char '.' 
                       ; es <- many digit
                       ; return $ (read (ds ++ "."++ es) :: Double)
                       }
                    <|> (return $ ((read ds) :: Double))
                  }

      
      skipLine = do newline
                    return ()
                    <|> do anyChar
                           skipLine
      
      
