--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Transform.hs
--
-- Author:
--   Klaus Aehlig (2010)
--
-- Description:
--    Compute the let- and sharing normal form of a term


module Transform where


import RAMLTypes
import Util
import List
import Control.Monad.State.Strict




--------------------------------------------------------------------------------
-- Unification of function declarations:
-- Every function should have exactly one argument!
-- Expressions should be in Share-Let-Normal-Form

normalizeFunDecs typedFunList = [ map normalizeDec funSCC | funSCC <- typedFunList ] -- list is in reverse topological order
   where
     normalizeDec (f,(context,e_f)) = 
         let (newVar:newVars) = [ ('^':f)++(show i) | i <- [1..] ]
             ef_normal = shareLetNormal newVars e_f
         in
         case context of
           [(x,t)] -> (f,(context, ef_normal))
           _       -> let (xs,ts) = unzip context
                          t = TTuple ts
                          e_f' = TEMatchT Free (TEVar newVar t) xs 
                                               (ef_normal) (typeOf e_f)
                      in (f,([(newVar,t)],e_f') )





--------------------------------------------------------------------------------
-- Substitutions


-- Substitute variables in arbritary expressions. 
subst :: TExp -> VarId -> VarId -> TExp
      -- e   ->  y    -> z     -> e[y:=z]

subst ex y z = sub ex
   where
     sub ex = 
         case ex of
           TEVar x t   -> TEVar (if x==y then z else x) t
           TEApp callName f e t -> TEApp callName f (sub e) t
           TEUnaryOp op e t -> TEUnaryOp op (sub e) t
           TEBinOp op e1 e2 t -> TEBinOp op (sub e1) (sub e2) t
           TESucc kind e t -> TESucc kind (sub e) t
           TEToInt e t -> TEToInt (sub e) t
           TEMatchN e e1 x e2 t -> TEMatchN (sub e) (sub e1) x (subBound [x] e2) t
           TELet kind x e1 e2 t -> TELet kind x (sub e1) (subBound [x] e2) t
           TECond e e1 e2 t -> TECond (sub e) (sub e1) (sub e2) t
           TECons kind e_h e_t t -> TECons kind (sub e_h) (sub e_t) t
           TETuple kind es t -> TETuple kind (map sub es) t
           TEMatchL mt e e1 x1 x2 e2 t -> TEMatchL mt (sub e) (sub e1) x1 x2 (subBound [x1,x2] e2) t
           TEMatchT mt e1 xs e2 t -> TEMatchT mt (sub e1) xs (subBound xs e2) t
           TENodeB kind e e1 e2 t -> TENodeB kind (sub e) (sub e1) (sub e2) t
           TEMatchB mt e e1 x1 x2 x3 e2 t -> TEMatchB mt (sub e) (sub e1) x1 x2 x3 (subBound [x1,x2,x3] e2) t
           TEShare x1 x2 x e t -> TEShare x1 x2 (if x==y then z else x) (subBound [x1,x2] e) t
           e -> e     -- nil, leaf, unit, true, false, ...

     subBound xs e = if y `elem` xs then e else sub e



-- Substitute variables along with freeLets in expressions.
-- This weird function is needed for a code transformation where
-- variables are replace by constructors.

-- subst :: TExp -> VarId -> VarId -> (TExp -> TExp) -> TExp
--       -- e   ->  y   -> z   -> freeLet  ->  something like e[y:=z]

substFreeLet ex y z freeLet = sub ex
   where
     sub ex = 
         case ex of
           TEVar x t   -> if x==y then freeLet (TEVar z t) else ex
           TEApp callName f e t -> TEApp callName f (sub e) t
           TEUnaryOp op e t -> TEUnaryOp op (sub e) t
           TEBinOp op e1 e2 t -> TEBinOp op (sub e1) (sub e2) t
           TESucc Normal e t -> TESucc Normal (sub e) t
           TESucc Free e t -> freeConstError
           TEToInt e t -> TEToInt (sub e) t
           TEMatchN e e1 x e2 t -> TEMatchN (sub e) (sub e1) x (subBound [x] e2) t
           TELet Normal x e1 e2 t -> TELet Normal x (sub e1) (subBound [x] e2) t
           TELet Free x e _ _ -> 
               let treatLikeAVariable = if y `elem` freeVars ex then 
                                            freeLet $ subst ex y z
                                        else ex    
               in
                 case e of
                   TENum Free _ _ -> treatLikeAVariable
                   TESucc Free _ _ -> treatLikeAVariable
                   TENil Free _ -> treatLikeAVariable
                   TECons Free _ _ _ -> treatLikeAVariable
                   TETuple Free _ _ -> treatLikeAVariable
                   TELeaf Free _ -> treatLikeAVariable
                   TENodeB Free _ _ _ _ -> treatLikeAVariable
                   _ -> error "Free let expression differs from the expected form."
           TECond e e1 e2 t -> TECond (sub e) (sub e1) (sub e2) t
           TECons Normal e_h e_t t -> TECons Normal (sub e_h) (sub e_t) t
           TECons Free e_h e_t t -> freeConstError
           TETuple Normal es t -> TETuple Normal (map sub es) t
           TETuple Free es t -> freeConstError
           TEMatchL mt e e1 x1 x2 e2 t -> TEMatchL mt (sub e) (sub e1) x1 x2 (subBound [x1,x2] e2) t
           TEMatchT mt e1 xs e2 t -> TEMatchT mt (sub e1) xs (subBound xs e2) t
           TENodeB Normal e e1 e2 t -> TENodeB Normal (sub e) (sub e1) (sub e2) t
           TENodeB Free e e1 e2 t -> freeConstError
           TEMatchB mt e e1 x1 x2 x3 e2 t -> TEMatchB mt (sub e) (sub e1) x1 x2 x3 (subBound [x1,x2,x3] e2) t
           TEShare x1 x2 x e t -> TEShare x1 x2 (if x==y then z else x) (subBound [x1,x2] e) t
           e -> e     -- nil, leaf, unit, true, false, ...

     freeConstError = error "Free constructor found in function substFreeLet."

     subBound xs e = if y `elem` xs then e else sub e


getVar :: State [String] String

getVar = do { nv:nvs <- get
            ; put nvs
            ; return nv
            }

getVars n = sequence [ getVar | i <- [1..n] ]


--------------------------------------------------------------------------------
-- Sharing, in CPS
--------------------------------------------------------------------------------


shareVar x e1 es cont = 
    do { nv1 <- getVar
       ; nv2 <- getVar
       ; result <- cont (subst e1 x nv1) (map (\e -> subst e x nv2) es)
       ; return $ TEShare nv1 nv2 x result (typeOf result)
       }

shareVars vars e1 es cont = case vars of
                              []     -> cont e1 es
                              (x:xs) -> shareVar x e1 es (\e1' -> \es' -> shareVars xs e1' es' cont)

shareVarsTuple es cont = 
    case es of
      [] -> cont []
      (e:es) -> shareVars (freeVars e `intersect` (foldl union [] (map freeVars es)) ) e es
                (\e' es' -> shareVarsTuple es' (\es'' -> cont $ e':es'') )



shareFreeVars e1 e2 cont = 
    shareVars ((freeVars e1) `intersect` (freeVars e2)) e1 [e2] (\e1' [e2'] -> cont e1' e2')



--------------------------------------------------------------------------------

freeLets lets exp = case lets of
                      [] -> return exp
                      (x,e):lets' -> do { exp' <- freeLets lets' exp
                                        ; return $ TELet Free x e exp' (typeOf exp')
                                        }
                   


--------------------------------------------------------------------------------


shareLetNormal :: [VarId] -> TExp -> TExp

shareLetNormal varNames e = evalState (norm e) varNames


norm exp = 
    if typeOf exp == TNatConst then fail $ "Error: Type TNatConst wasn't eliminated!\n In Expression \n" ++ (show exp)  -- XXX sanity check
    else
    case exp of
       TEApp callName f e t ->       -- do we need a fresh variable here ???
          do { e' <- norm e
             ; return $ TEApp callName f e' t
             }

       TEUnaryOp op e t ->
          do { e' <- norm e
             ; return $ TEUnaryOp op e' t
             }

       TEBinOp b e1 e2 t ->
          do { e1' <- norm e1
             ; e2' <- norm e2
             ; [nv1,nv2] <- getVars 2
             ; shareFreeVars e1' e2' (\ e1' e2' -> 
               freeLets [(nv1,e1'),(nv2,e2')]
               (TEBinOp b (TEVar nv1 (typeOf e1')) (TEVar nv2 (typeOf e2')) t) )
             }

       TESucc Normal e t -> 
          do { e' <- norm e
             ; nv <- getVar
             ; freeLets [(nv,e')] $ TESucc Normal (TEVar nv (typeOf e')) t
             }

       TESucc Free e t -> error "Trying to transform a free succ."

       TEToInt e t -> 
          do { e' <- norm e
             ; nv <- getVar
             ; freeLets [(nv,e')] $ TEToInt (TEVar nv (typeOf e')) t
             }

       TEMatchN e e1 x e2 t ->
           do { e' <- norm e
              ; e1' <- case e of 
                         TEVar y TNat | y `elem` (freeVars e1)  -> 
                             do { ny <- getVar
                                ; norm $ substFreeLet e1 y ny (\e -> TELet Free ny (TENum Free 0 TNat) e (typeOf e))
                                }
                         _ -> norm e1
              ; e2' <- case e of
                         TEVar y _ | y == x -> norm e2   -- Watch out with shadowing of y
                         TEVar y TNat | y `elem` (freeVars e2) -> 
                             do { ny <- getVar
                                ; let expY = TESucc Free (TEVar x TNat) TNat
                                ; let ne2 = substFreeLet e2 y ny (\e -> TELet Free ny expY e (typeOf e))
                                -- Check whether some pattern matches along the way to y shadow x!
                                ; if x `elem` (freeVars ne2) then
                                      norm ne2
                                  else -- Don't do the replacement if x1 or x2 is shadowed 
                                      norm e2
                                }
                         _ -> norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchN (TEVar nv (typeOf e')) e1' x e2' t)
              } 

       TELet Normal x e1 e2 t ->
           do { e1' <- norm e1
              ; e2' <- norm e2
              ; let commonVars = ((freeVars e1') `intersect` ((freeVars e2) \\ [x]))
              ; shareVars commonVars e1' [e2'] (\ e1' [e2'] ->
                return $ TELet Normal x e1' e2' t )
              }

       TELet Free x e _ _ -> 
           case e of
             TENum Free _ _ -> return exp
             TESucc Free _ _ -> return exp
             TENil  Free _   -> return exp
             TECons Free _ _ _ -> return exp
             TETuple Free _ _ -> return exp
             TELeaf Free _    -> return exp
             TENodeB Free _ _ _ _ -> return exp
             _ -> error $ "Free let expression differs from the expected form in the function 'norm'. \n\n" ++ (show exp)

       TEShare x1 x2 x e t ->
           do { e' <- norm e 
              ; return $ TEShare x1 x2 x e' t
              }

       TECond e e1 e2 t ->
           do { e' <- norm e
              ; e1' <- norm e1
              ; e2' <- norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` ((freeVars e1') `union` (freeVars e2'))
              ; shareVars commonVars e' [e1',e2'] (\e' [e1',e2'] ->
                freeLets [(nv,e')] $ TECond (TEVar nv (typeOf e')) e1' e2' t )
              }

       TECons Normal e1 e2 t ->
           do { e1' <- norm e1
              ; e2' <- norm e2
              ; [nv1,nv2] <- getVars 2
              ; shareFreeVars e1' e2' (\ e1' e2' -> 
                  freeLets [(nv1,e1'),(nv2,e2')] $
                     TECons Normal (TEVar nv1 (typeOf e1')) (TEVar nv2 (typeOf e2')) t )
              }

       TECons Free _ _ _ -> error "Trying to transform a free cons."

       TEMatchL mt e e1 x1 x2 e2 t ->
           do { e' <- norm e
              ; e1' <- case e of 
                         TEVar y listT | y `elem` (freeVars e1)  -> 
                             do { ny <- getVar
                                ; norm $ substFreeLet e1 y ny (\e -> TELet Free ny (TENil Free listT) e (typeOf e))
                                }
                         _ -> norm e1
              ; e2' <- case e of
                         TEVar y _ | y `elem` [x1,x2] -> norm e2   -- Watch out with shadowing of y
                         TEVar y listT@(TList elemT) | y `elem` (freeVars e2) -> 
                             do { ny <- getVar
                                ; let expY = TECons Free (TEVar x1 elemT) (TEVar x2 listT) listT
                                ; let ne2 = substFreeLet e2 y ny (\e -> TELet Free ny expY e (typeOf e))
                                -- Check whether some pattern matches along the way to y shadow x1 or x2!
                                ; if [x1,x2] `intersect` (freeVars ne2) == [x1,x2] then
                                      norm ne2
                                  else -- Don't do the replacement if x1 or x2 is shadowed 
                                      norm e2
                                }
                         _ -> norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x1,x2]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchL mt (TEVar nv (typeOf e')) e1' x1 x2 e2' t)
              } 

       TETuple Normal es t ->
           do { es' <- sequence [ norm e | e <- es]
              ; nvs <- getVars (length es)
              ; shareVarsTuple es' (\ es' ->
                let varsEs = zip nvs es' in
                freeLets varsEs  $ TETuple Normal (map (\(nv,e) -> TEVar nv (typeOf e)) varsEs ) t )
              }

       TETuple Free _ _ -> error "Trying to transform a free tuple."

       TEMatchT mt e xs e1 t ->
           do { e' <- norm e
              ; e1'  <- case e of
                         TEVar y _ | y `elem` xs -> norm e1   -- Watch out with shadowing of y
                         TEVar y tupleT@(TTuple elemTs) | y `elem` (freeVars e1) -> 
                             do { ny <- getVar
                                ; let varExps = map ( \(x,t) -> TEVar x t ) (zip xs elemTs)
                                ; let expY = TETuple Free varExps tupleT
                                ; let ne1 = substFreeLet e1 y ny (\e -> TELet Free ny expY e (typeOf e))
                                -- Check whether some pattern matches along the way to y shadows one of the xs!
                                ; if xs `intersect` (freeVars ne1) == xs then
                                      norm ne1
                                  else -- Don't do the replacement if an element of xs is shadowed 
                                      norm e1
                                }
                         _ -> norm e1
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` ((freeVars e1') \\ xs)
              ; shareVars commonVars e' [e1'] (\ e' [e1'] ->
                freeLets [(nv,e')] $ TEMatchT mt (TEVar nv (typeOf e')) xs e1' t )
              }
 
       TENodeB Normal e1 e2 e3 t ->
           do { es' <- sequence [ norm e | e <- [e1,e2,e3] ]
              ; nvs <- getVars 3
              ; shareVarsTuple es' (\ es' ->
                let varsEs = zip nvs es'
                    [v1,v2,v3] = map (\(nv,e) -> TEVar nv (typeOf e)) varsEs
                in
                freeLets varsEs  $ TENodeB Normal v1 v2 v3 t )
              }

       TENodeB Free _ _ _ _ -> error $ "Trying to transfor a free node."

       TEMatchB mt e e1 x1 x2 x3 e2 t ->
           do { e' <- norm e
              ; e1' <- case e of
                         TEVar y treeT | y `elem` (freeVars e1)  -> 
                             do { ny <- getVar
                                ; norm $ substFreeLet e1 y ny (\e -> TELet Free ny (TELeaf Free treeT) e (typeOf e))
                                }
                         _ -> norm e1
              ; e2' <- case e of
                         TEVar y _ | y `elem` [x1,x2,x3] -> norm e2   -- Watch out with shadowing of y
                         TEVar y treeT@(TList elemT) | y `elem` (freeVars e2) ->  
                             do { ny <- getVar
                                ; let expY = TENodeB Free (TEVar x1 elemT) (TEVar x2 elemT) (TEVar x3 treeT) treeT
                                ; let ne2 = substFreeLet e2 y ny (\e -> TELet Free ny expY e (typeOf e))
                                -- Check whether some pattern matches along the way to y shadow x1, x2 or x3!
                                ; if [x1,x2,x3] `intersect` (freeVars ne2) == [x1,x2,x3] then
                                      norm ne2
                                  else -- Don't do the replacement if x1, x2 or x3 is shadowed 
                                      norm e2
                                }
                         _ -> norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x1,x2,x3]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchB mt (TEVar nv (typeOf e')) e1' x1 x2 x3 e2' t)
              } 

--  TEUnit TETrue TEFalse TENat TEVar TENil TELeaf TETick
       _ -> return exp

